import inspect
import os
import re
from copy import copy
from pathlib import Path
from typing import (
    IO,
    Any,
    Union,
)

from bs4 import BeautifulSoup

from .env import env

html_regex = r"<!.*?->|^--html"

project_root = os.getcwd()

_cached_html: dict[str, Any] = {}


def load_html_to_soup(path: Union[str, Path], parser: str = "html.parser"):
    """
    Load an HTML file from a specified path and return a BeautifulSoup object.
    The function automatically determines the base path using the caller's file path.

    :param path: The path to the HTML file.
    :return: A BeautifulSoup object of the loaded HTML file.
    """
    if isinstance(path, BeautifulSoup):
        return copy(path)

    # If the path is multiple lines or starts with --html, it contains the HTML
    if isinstance(path, str) and ("\n" in path or path.startswith("--html")):
        file_contents = re.sub(html_regex, "", path, flags=re.DOTALL)

        first_line = file_contents.split("\n")[0]

        if "<?xml" in first_line:
            parser = "xml"

        return BeautifulSoup(file_contents, parser)

    if "~/" in str(path) or bool(re.match(r"^[a-zA-Z]", str(path))):
        path = Path(str(path).replace("~/", ""))

        base_path = project_root
    else:
        # Convert the path to a `Path` object if necessary
        if not isinstance(path, Path):
            path = Path(path)

        # Traverse the stack to find the first frame outside of the 'weba' package
        for frame_info in inspect.stack():
            frame_file = Path(frame_info.filename)

            if "weba" not in frame_file.parts:
                # Use the caller's file path to determine the base path
                base_path = frame_file.parent
                break
            else:
                base_path = None
        else:
            # If no caller is found outside of 'weba', default to the current working directory
            base_path = Path.cwd()

    # Resolve the path relative to the base path
    resolved_path = base_path / path

    # Ensure the file exists
    if not resolved_path.exists():
        raise FileNotFoundError(f"The file at {resolved_path} does not exist.")

    global _cached_html

    if env.is_prd:
        if cached_html := _cached_html.get(str(resolved_path)):
            return BeautifulSoup(cached_html, parser)

    # Read the file and parse it with BeautifulSoup
    with open(resolved_path, "r", encoding="utf-8") as file:
        return _load_file_contents_to_bs4(file, parser, resolved_path)


def _load_file_contents_to_bs4(file: IO[str], parser: str, resolved_path: Path) -> BeautifulSoup:
    global _cached_html

    file_contents = file.read()

    # Remove comments
    file_contents = re.sub(html_regex, "", file_contents, flags=re.DOTALL)

    first_line = file_contents.split("\n")[0]

    if "<?xml" in first_line:
        parser = "xml"

    _cached_html[str(resolved_path)] = file_contents

    return BeautifulSoup(file_contents, parser)
